"""
interval.ddd.event
~~~~~~~~~~~~~~~~~~

This module provides DDD event base classes.
"""

import dataclasses
import datetime
import uuid


@dataclasses.dataclass
class DomainEvent:
    """领域事件

    Attributes:
        id: 事件ID
        occurred_at: 事件发生时间（包含系统本地时区）
    """
    id: str = dataclasses.field(
        default_factory=lambda: str(uuid.uuid1()),
        init=False
    )
    occurred_at: datetime.datetime = dataclasses.field(
        default_factory=lambda: datetime.datetime.now().astimezone(),
        init=False
    )



"""
interval.utils
~~~~~~~~~~~~~~

This module provides utility functions.
"""

from decimal import Decimal
from logging import Filter, Formatter, getLogger, Logger, LogRecord, StreamHandler
import random
import re
import string
from typing import Any, Callable

import orjson


def get_stream_logger(name: str, level: int | str,
                      filters: list[Filter | Callable[[LogRecord], Any]] = None,
                      formatter: Formatter = None) -> Logger:
    """获取日志记录器（日志输出到sys.stderr）

    Args:
        name: 日志名称
        level: 日志级别
        filters: logging.Filter实例（或者以logging.LogRecord实例作为参数的函数）列表
        formatter: logging.Formatter实例

    Returns:
        日志记录器
    """
    logger = getLogger(name)
    logger.setLevel(level)
    if filters:
        for f in filters:
            logger.addFilter(f)
    handler = StreamHandler()
    handler.setLevel(level)
    if formatter is not None:
        handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def generate_nonce(length: int, chars: str = 'uld', population: str = '',
                   *, prefix: str = '', suffix: str = '') -> str:
    """生成固定长度的随机字符串

    Args:
        length: 字符串总长度
        chars: 使用的字符种类：'u' - 大写字母，'l' - 小写字母，'d' - 数字
        population: 使用的字符集合；若指定了population则忽略chars
        prefix: 字符串前缀
        suffix: 字符串后缀

    Returns:
        随机字符串

    Raises:
        ValueError: 前缀和后缀的长度之和大于总长度；没有指定有效的字符种类或者字符集合
    """
    random_part_len = length - len(prefix + suffix)
    if random_part_len < 0:
        raise ValueError('Invalid length')
    if random_part_len == 0:
        return prefix + suffix
    if not population:
        if 'u' in chars:
            population += string.ascii_uppercase
        if 'l' in chars:
            population += string.ascii_lowercase
        if 'd' in chars:
            population += string.digits
        if not population:
            raise ValueError('Invalid chars')
    elements = random.choices(population, k=random_part_len)
    return prefix + ''.join(elements) + suffix


def check_mobile_number(number: str) -> bool:
    """检查手机号码格式是否正确

    Args:
        number: 手机号码

    Returns:
        格式正确返回True，否则返回False
    """
    pattern = r'1[3-9]\d{9}'
    match_obj = re.fullmatch(pattern, number)
    return bool(match_obj)


def check_id_card_number(number: str) -> bool:
    """检查身份证号码格式是否正确

    Args:
        number: 身份证号码

    Returns:
        格式正确返回True，否则返回False
    """
    if len(number) == 18:
        pattern = r'[1-9]\d{5}[12]\d{3}(0[1-9]|1[0-2])(0[1-9]|[12]\d|3[01])\d{3}[\dxX]'
        match_obj = re.fullmatch(pattern, number)
        return bool(match_obj)
    if len(number) == 15:
        pattern = r'[1-9]\d{7}(0[1-9]|1[0-2])(0[1-9]|[12]\d|3[01])\d{3}'
        match_obj = re.fullmatch(pattern, number)
        return bool(match_obj)
    return False


def _orjson_default(obj: Any) -> Any:
    if isinstance(obj, set):
        return list(obj)
    if isinstance(obj, Decimal):
        return str(obj)
    raise TypeError


def orjson_dumps(obj: Any) -> bytes:
    """使用orjson进行JSON序列化操作，支持set与Decimal类型

    Args:
        obj: Python对象

    Returns:
        JSON字节流

    Raises:
        orjson.JSONEncodeError: TypeError的子类
    """
    return orjson.dumps(
        obj,
        default=_orjson_default,
        option=orjson.OPT_OMIT_MICROSECONDS | orjson.OPT_SERIALIZE_NUMPY
    )


def orjson_loads(obj: bytes | str) -> Any:
    """使用orjson进行JSON反序列化操作

    Args:
        obj: JSON字节流或字符串

    Returns:
        Python对象

    Raises:
        orjson.JSONDecodeError: ValueError的子类
    """
    return orjson.loads(obj)



"""
interval.ddd.entity
~~~~~~~~~~~~~~~~~~~

This module provides DDD Entity base class and derived classes.
"""


class Entity:
    """实体

    Attributes:
        ref: 唯一标识
    """

    def __init__(self, ref):
        self.ref = ref

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return other.ref == self.ref

    def __hash__(self):
        return hash(self.ref)

    def __repr__(self):
        return f'<{type(self).__name__} {self.ref!r}>'


class Aggregate(Entity):
    """聚合根实体

    Attributes:
        ref: 唯一标识
        version_number: 版本号，用于支持乐观锁
        domain_events: 领域事件列表
    """

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        obj.domain_events = []
        return obj

    def __init__(self, ref, version_number: int = 1):
        super().__init__(ref)
        self.version_number = version_number



"""
interval.ddd.valueobject
~~~~~~~~~~~~~~~~~~~~~~~~

This module provides DDD ValueObject base class and derived classes.
"""

import dataclasses
import datetime
import typing
import uuid

try:
    from bson import ObjectId
except ImportError:
    ObjectId = None


@dataclasses.dataclass(frozen=True)
class ValueObject:
    """值对象"""

    @classmethod
    def composite_factory(cls, *args) -> typing.Self | None:
        """一个工厂方法，用于支持SQLAlchemy ORM的“复合列类型”，对于“嵌套复合”不适用"""
        for arg in args:
            if arg is not None:
                return cls(*args)  # noqa


@dataclasses.dataclass(frozen=True)
class IntegerRef(ValueObject):
    """唯一标识（整数）"""
    value: int


@dataclasses.dataclass(frozen=True)
class StringRef(ValueObject):
    """唯一标识（字符串）"""
    value: str


def _gen_uuid_str() -> str:
    return uuid.uuid1().hex


@dataclasses.dataclass(frozen=True)
class UUIDRef(StringRef):
    """唯一标识（UUID）"""
    value: str = dataclasses.field(default_factory=_gen_uuid_str)

    @property
    def typed_value(self) -> uuid.UUID:
        return uuid.UUID(self.value)

    @property
    def created_at(self) -> datetime.datetime:
        """标识创建时间（包含系统本地时区）"""
        t = (self.typed_value.time - 0x01b21dd213814000) / 10_000_000
        return datetime.datetime.fromtimestamp(t).astimezone()


def _gen_oid_str() -> str:
    if ObjectId is None:
        raise RuntimeError('PyMongo package is not installed')
    return str(ObjectId())


@dataclasses.dataclass(frozen=True)
class OIDRef(StringRef):
    """唯一标识（ObjectId）"""
    value: str = dataclasses.field(default_factory=_gen_oid_str)

    @property
    def typed_value(self) -> ObjectId:
        if ObjectId is None:
            raise RuntimeError('PyMongo package is not installed')
        return ObjectId(self.value)

    @property
    def created_at(self) -> datetime.datetime:
        """标识创建时间（包含系统本地时区）"""
        return self.typed_value.generation_time.astimezone()



"""
interval.ddd.exceptions
~~~~~~~~~~~~~~~~~~~~~~~

This module provides DDD exceptions.
"""

import functools
import sys


class DDDException(Exception):
    """根异常"""
    code: int = None  # 错误码
    msg: str = None  # 错误描述


class DomainException(DDDException):
    """领域异常"""
    pass


class ServiceLayerException(DDDException):
    """服务层异常"""
    pass


class AdapterException(DDDException):
    """适配器异常"""
    pass


class RemoteServiceException(AdapterException):
    """远程服务异常"""
    pass


class DBAPIError(AdapterException):
    """Wraps a DB-API 2.0 Error"""

    @property
    def orig(self) -> Exception | None:
        """Original exception"""
        return self.__cause__

    def __str__(self):
        if self.orig:
            return self.orig.__str__()
        else:
            return super().__str__()


class InterfaceError(DBAPIError): pass
class DatabaseError(DBAPIError): pass
class DataError(DatabaseError): pass
class OperationalError(DatabaseError): pass
class IntegrityError(DatabaseError): pass
class InternalError(DatabaseError): pass
class ProgrammingError(DatabaseError): pass
class NotSupportedError(DatabaseError): pass


class DBAPIErrorWrapper:
    """DBAPIError包装器

    将数据库适配器抛出的异常转换成相应的统一定义的DBAPIError；作为装饰器或上下文管理器使用。

    Attributes:
        errors: 异常名称与DBAPIError类型的对应关系；默认使用STANDARD_DBAPI_ERRORS
        default: 如果没有在errors中找到对应关系，则将异常转换成该属性指定的DBAPIError类型；默认为None，不进行默认转换
    """

    def __init__(self,
                 errors: dict[str, type[DBAPIError]] = None,
                 default: type[DBAPIError] = None):
        if errors is None:
            self.errors = STANDARD_DBAPI_ERRORS
        else:
            self.errors = errors
        self.default = default

    def raise_from(self, exc_type, exc_val, exc_tb):
        if exc_type.__name__ in self.errors:
            new_type = self.errors[exc_type.__name__]
        elif self.default:
            new_type = self.default
        else:
            return
        new_exc = new_type(*exc_val.args)
        if exc_tb:
            raise new_exc.with_traceback(exc_tb) from exc_val
        else:
            raise new_exc from exc_val

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            return
        self.raise_from(exc_type, exc_val, exc_tb)

    def __call__(self, func):
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception:
                self.raise_from(*sys.exc_info())
                raise
        return _wrapper


STANDARD_DBAPI_ERRORS = {
    'InterfaceError': InterfaceError,
    'DatabaseError': DatabaseError,
    'DataError': DataError,
    'OperationalError': OperationalError,
    'IntegrityError': IntegrityError,
    'InternalError': InternalError,
    'ProgrammingError': ProgrammingError,
    'NotSupportedError': NotSupportedError
}



"""
interval
~~~~~~~~

Utility functions and classes used in applications of Interval Design.

:copyright: (c) 2022 by Interval Design.
:license: MIT, see LICENSE for more details.
"""

from . import ddd
from . import utils



"""
interval.ddd
~~~~~~~~~~~~

This package provides basic components of Domain-Driven Design.
"""

from .dto import UseCaseDTO
from .entity import Entity, Aggregate
from .event import DomainEvent
from .exceptions import (
    DDDException,
    DomainException,
    ServiceLayerException,
    AdapterException,
    RemoteServiceException,
    DBAPIError,
    InterfaceError,
    DatabaseError,
    DataError,
    OperationalError,
    IntegrityError,
    InternalError,
    ProgrammingError,
    NotSupportedError,
    DBAPIErrorWrapper,
    STANDARD_DBAPI_ERRORS
)
from .valueobject import (
    ValueObject,
    IntegerRef,
    StringRef,
    UUIDRef,
    OIDRef
)
